package entities;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Set;
import java.util.Map.Entry;

import utilities.Utils;

public class Section {
	private List<String> title = new ArrayList<String>();
	private List<List<String>> sentences = new ArrayList<List<String>>();
	private List<Section> subsections = new ArrayList<Section>();
	private Set <String> allBodyWords = new HashSet<String>();
	private Hashtable<String, Integer> bodyWordFrequencies = new Hashtable<String, Integer>();
	private Hashtable<String, Integer> bodyWordsFirstSentenceAppearance = new Hashtable<String, Integer>();
	private Hashtable<String, Integer> word2countRank = new Hashtable<String, Integer>();
	private int numberTotalWords = 0;
	private Section supersection = null;
	private boolean isIgnore = false;
	
	/**
	 * Creates a Section Object
	 * @param title - Title of the section
	 */
	public Section(String title) {
		super();
		this.title = Utils.parseIntoWords(title);
	}

	/**
	 * Creates a Section Object
	 * @param title - Title of the section
	 */
	public Section(String title, boolean isIgnore) {
		super();
		this.title = Utils.parseIntoWords(title);
		this.isIgnore = isIgnore;
	}
	
	/**
	 * Adds a sentence to the section.  Sentences must be added in reading order.
	 * @param sentence Sentence to add
	 */
	public void addSentence(String sentence) {
		this.addSentence(Utils.parseIntoWords(sentence));
	}
	
	/**
	 * Internal method that adds sentences
	 * @param sentenceWords Sentence to add
	 */
	private void addSentence(List<String> sentenceWords) {
		// Add to list of sentences
		sentences.add(sentenceWords);
		
		for(String word : sentenceWords) {
			// Add to body words
			allBodyWords.add(word);
			
			// Update first sentence appearance
			if(!bodyWordsFirstSentenceAppearance.containsKey(word)) {
				bodyWordsFirstSentenceAppearance.put(word, sentences.size());
			}

			// Update word frequencies
			int wordFreq = 1;
			if(bodyWordFrequencies.containsKey(word)) {
				wordFreq += bodyWordFrequencies.get(word);
			}
			bodyWordFrequencies.put(word, wordFreq);
		}
		
		// Update total number of words count
		this.numberTotalWords += sentenceWords.size();
		
		// Calculate Frequency Rank of all words in this section
		word2countRank.clear();
		//	Add them to word => count table
		Hashtable<String, Integer> word2count = new Hashtable<String, Integer>();
		for(String word : this.allBodyWords) {
			if(!word2count.containsKey(word)) {
				word2count.put(word, this.getWordOccurences(word));
			}
		}
		
		// Extract the counts without duplicates
		HashSet<Integer> wordCountValuesSet = new HashSet<Integer>();
		wordCountValuesSet.addAll(word2count.values());
		ArrayList<Integer> wordCountValues = new ArrayList<Integer>();
		wordCountValues.addAll(wordCountValuesSet);
		
		// Sort them in ascending order
		Collections.sort(wordCountValues);
		Collections.reverse(wordCountValues);
		
		// Add to word => count rank table
		for(Entry<String, Integer> entry : word2count.entrySet()) {
			word2countRank.put(entry.getKey(), wordCountValues.indexOf(entry.getValue()));
		}
	}
	
	/**
	 * Get the index at which this word first appears (1 based)
	 * @param word Word to search for
	 * @return  index of the word
	 */
	public int getFirstWordIndex(String word) {
		// Get the index of the first sentence it appeared in.
		int firstSentenceIndex = bodyWordsFirstSentenceAppearance.get(word)-1;
		
		// Find one which word in the sentence it occured
		int index = this.sentences.get(firstSentenceIndex).indexOf(word)+1;
		
		// Add the number of words in previous sentences
		for(int i=0; i < firstSentenceIndex; i++) {
			index += this.sentences.get(i).size();
		}
		return index;
	}

	// Returns list of all sentences in this section
	public List<List<String>> getSentences() {
		return this.sentences;
	}
	
	/**
	 * Returns true if this body of this section contains the specified word
	 */
	public boolean bodyContainsWord(String word) {
		return allBodyWords.contains(word);
	}
	
	/**
	 * Returns the index of the sentence in which this word first appears
	 * @param word
	 * @return
	 */
	public int getFirstSentenceIndexWithWord(String word) {
		return this.bodyWordsFirstSentenceAppearance.get(word);
	}
	
	/** Returns true if the specified word is in the section title **/ 
	public boolean isWordInTitle(String word) {
		return this.title.contains(word);
	}
	
	/** Returns true if the specified word is in the section body **/
	public boolean isWordInBody(String word) {
		return this.allBodyWords.contains(word);
	}
	
	/** Returns the frequency of this word in the section **/
	public int getWordOccurences(String word) {
		return this.bodyWordFrequencies.get(word);
		
	}
	
	/** Returns the total number of words in the body of this section not including the title **/
	public int getNumberOfTotalBodyWords() {
		return this.numberTotalWords;
	}
	
	/** Returns the total number of unique words in the section not including the title **/ 
	public int getNumberOfTotalUniqueBodyWords() {
		return this.allBodyWords.size();
	}
	
	/** Returns the frequency rank of the word within the section.  
	 * The lower the rank number the more the word appears in the section **/ 
	public int getFrequencyRank(String word) {
		return word2countRank.get(word);
	}
	
	/** Adds the specified section as subsection of this section **/
	public void addSubsection(Section subsection) {
		this.subsections.add(subsection);
		
		// Add this as parent of subsection
		subsection.setParent(this);
		
		// Add subsection text to sentences
		//for(List<String> sentence : subsection.getSentences()) {
		//	this.addSentence(sentence);
		//}
	}
	
	/** Returns the subsections of this section **/
	public List<Section> getSubsections() {
		return this.subsections;
	}
	
	/** Sets the specified section as a parent of this section **/
	private void setParent(Section parentSection) {
		this.supersection = parentSection;
	}
	
	/** Returns the parent section of this section **/
	public Section getParent() {
		return this.supersection;
	}
	
	/** Returns how many levels deep this node is **/
	public int getDepth() {
		int depth = 0;
		Section parentSection = this;
		while(parentSection != null) {
			parentSection = parentSection.getParent();
			depth++;
		}
		return depth;
	}
	
	public String getTitleString() {
		String titleString = "";
		for(String tString : this.title) {
			titleString += tString+" ";
		}
		return titleString;
	}
	
	public String toString() {
		return this.title.toString()+" "+this.sentences;
	}
	
	public boolean wordInParentBody(String word) {
		return this.getParent() != null && this.getParent().isWordInBody(word);
	}
	
	public boolean isIgnore() {
		return this.isIgnore;
	}
	
	public boolean wordInAnySiblingBody(String word) {
		if (this.getParent() !=null) {
			for(Section sibling : this.getParent().getSubsections()) {
				if(sibling != this) {
					if(sibling.isWordInBody(word)) {
						return true;
					}
				}
			}
		}
		return false;
	}
	
	public boolean wordInAnySiblingTitle(String word) {
		if(this.getParent() !=null) {
			for(Section sibling : this.getParent().getSubsections()) {
				if(sibling != this) {
					if(sibling.isWordInTitle(word)) {
						return true;
					}
				}
			}
		}
		return false;
	}
}
