import colorset, node, score_combiner, config, beam_queue, utils, new_timer
import random, sys, math

class Graph:

    def __init__(self, stemmed_section_word_scores, stem_probs, lang_model,\
                 stopwords, tag_probs, word_to_tag, terminal_word_probs, section_number):

        self.nodes = []
        self.node_groups_by_stem = {}
        self.edge_trans = []
        self.stopwords = stopwords
        
        # Decouple terminal word probs
        (start_word_probs, end_word_probs) = terminal_word_probs
        
        # Add all nodes to graph: One node per unstemmed word of each
        # stemmed word in the section
        for stemmed_section_word in stemmed_section_word_scores.keys():

            # If this word cannot be destemmed, skip
            if not stem_probs.has_key(stemmed_section_word):
                continue
            
            # Create new node stem group
            self.node_groups_by_stem[stemmed_section_word] = []
            
            unstemmed_section_words = stem_probs[stemmed_section_word].keys()
            for unstemmed_section_word in unstemmed_section_words:
                
                # Create and add new node to graph
                new_node = node.Node(unstemmed_section_word, stemmed_section_word, len(self.nodes))
                self.add_node(new_node)

                # Add node to group of node with same stem
                self.node_groups_by_stem[stemmed_section_word].append(new_node)

        # Initialize edge trans
        for node1_index in range(len(self.nodes)+2): # +2 for start/end
            self.edge_trans.append([])
            for node2_index in range(len(self.nodes)+2): # +2 for start/end
                self.edge_trans[node1_index].append([])
                for stopword_index in range(len(self.stopwords)+1): # +1 for direct edge
                    self.edge_trans[node1_index][node2_index].append(None)
                    
        # Add edge from each node to all other nodes using bigram
        # probabilities and word selection scores
        total_nodes_str = str(len(self.nodes))
        for from_node_num, from_node in enumerate(self.nodes):

            print str(from_node_num)+" of "+total_nodes_str
            sys.stdout.flush()

            for to_node in self.nodes:

                # Get from node selection score
                selection_score = stemmed_section_word_scores[to_node.stemmed_word]

                if from_node != to_node:
                    
                    from_node_tag = word_to_tag.get(from_node.word, section_number)
                    to_node_tag = word_to_tag.get(to_node.word, section_number)
                    
                    # Calculate transition scores between this word
                    # and the next word using stop words
                    for stopword_index, stopword in enumerate(stopwords):

                        stopword_tag = word_to_tag.get(stopword, section_number)

                        # Calculate transition probability from first word to stop word
                        first_word_prob = lang_model.get(from_node.word, stopword)
                        first_tag_prob = tag_probs.get(from_node_tag, stopword_tag)

                        # Calculate transition probability from stop word to second word
                        second_word_prob = lang_model.get(stopword, to_node.word)
                        second_tag_prob = tag_probs.get(stopword_tag, to_node_tag)

                        # Combine scores
                        word_prob = first_word_prob * second_word_prob

                        tag_prob = first_tag_prob * second_tag_prob
                        prob = score_combiner.combine_selection_bigram_tag_scores(\
                            selection_score, word_prob, tag_prob)

                        # Add edge
                        self.add_edge(from_node, to_node, stopword_index, prob)
                        
                    # Calculate the bigram score with no stop words
                    word_prob = lang_model.get(from_node.word, to_node.word)

                    # Calculate the tag tranisition probability with no stop words
                    tag_prob = tag_probs.get(from_node_tag, to_node_tag)

                    # Combine the two probabilities
                    prob =  score_combiner.combine_selection_bigram_tag_scores(\
                            selection_score, word_prob, tag_prob)

                    # Add edge
                    self.add_edge(from_node, to_node, config._NO_TRANS_INDEX, prob)
                    
        # Add special start node connected to all word nodes
        self.start_node = node.Node("#START#", "#START#", len(self.nodes))
        
        # Connect start node
        for to_node in self.nodes:

            # Get scores for each part
            start_selection_score = stemmed_section_word_scores[to_node.stemmed_word]

            start_word_prob = start_word_probs.get(to_node.word, 0.0)

            start_word_prob = 1.0
            
            start_tag_prob = tag_probs.get("#START#", word_to_tag.get(to_node.word, section_number))

            # Add edge
            prob = score_combiner.combine_selection_bigram_tag_scores(\
                start_selection_score, start_word_prob, start_tag_prob)
            self.add_edge(self.start_node, to_node, config._NO_TRANS_INDEX, prob)
        self.add_node(self.start_node)

        # Add special end node connected from all word nodes
        self.end_node = node.Node("#END#", "#END#", len(self.nodes))
                        
        # Connect end node
        for from_node in self.nodes:

            # There should be no edge from the start node to the end node
            if from_node == self.start_node:
                continue

            # Get scores for each part
            end_selection_score = 1.0

            end_word_prob = end_word_probs.get(from_node.word, 0.0)

            end_word_prob = 1.0

            end_tag_prob = tag_probs.get(word_to_tag.get(from_node.word, section_number), "#END#")

            # Add edge
            prob = score_combiner.combine_selection_bigram_tag_scores(\
                end_selection_score, end_word_prob, end_tag_prob)
            self.add_edge(from_node, self.end_node, config._NO_TRANS_INDEX, prob)
        self.add_node(self.end_node)

    def add_edge(self, from_node, to_node, stopword_index, prob):
        if stopword_index == config._NO_TRANS_INDEX:
            stopword_index = len(self.stopwords)
        self.edge_trans[from_node.number][to_node.number][stopword_index] = math.log( prob )
        
    ''' Adds a node to this graph '''
    def add_node(self, node):
        self.nodes.append(node)

    def phrase_from_edge_and_node_list(self, edge_list, node_list):

        assert len(node_list) == len(edge_list)+1
        
        output_phrase = ""
        
        for i in range(len(node_list)-1):
            stopword = ( edge_list[i] == len(self.stopwords) or edge_list[i] == config._NO_TRANS_INDEX) and " " \
                       or self.stopwords[edge_list[i]]            
            output_phrase += node_list[i].word+" "+stopword+" "
        output_phrase += node_list[-1].word
        
        return output_phrase

    def beam_search(self, max_path_length, beam_size):

        t_timer = new_timer.Timer()

        beam_q = beam_queue.BeamQueue(beam_size)

        start_path_score = self.__calc_iterative_score(0.0, 0.0)
        beam_q.add( tuple(), self.start_node, tuple(), None, start_path_score )

        # Loop for the whole path length
        for path_length in range(max_path_length+1):

            # Make a new beam queue
            new_beam_q = beam_queue.BeamQueue(beam_size)

            # Loop through each path in the beam
            for (phrase_nodes, phrase_edges), phrase_score in beam_q.items():

                # Get the last node in the path
                last_node = phrase_nodes[-1]

                # Explore all to nodes
                for to_node in self.nodes:

                    # Explore all edges to the to node
                    for edge_stopword_index, edge_score in enumerate(self.edge_trans[last_node.number][to_node.number]):
                        if edge_score == None:
                            continue

                        # Check for redundant stems
                        has_redundant_stem = False
                        for phrase_node in phrase_nodes:
                            if phrase_node.stemmed_word == to_node.stemmed_word:
                                has_redundant_stem = True
                                break
                        if has_redundant_stem:
                            continue
                        
                        # Don't end with the end node
                        # If this is the last node in the path, make sure it ends with the end node
                        if (to_node == self.end_node and path_length < max_path_length) or \
                               (to_node != self.end_node and path_length == max_path_length):
                            continue

                        new_phrase_score = self.__calc_iterative_score(phrase_score, edge_score)

                        # Add to new beam queue
                        new_beam_q.add( phrase_nodes, to_node, phrase_edges, edge_stopword_index, new_phrase_score)

            # Set the new beam
            beam_q = new_beam_q

        # Return top result in beam
        ((top_phrase_nodes, top_phrase_edges), top_phrase_score) = beam_q.top_element()
        total_time = t_timer.total()
        
        return (top_phrase_score, top_phrase_edges, top_phrase_nodes, total_time)

    def decolor(self):
        for node in self.nodes:
            node.color = -1
            node.color_sets = {}
        
    def randomized_search(self, k, r, node_groups, all_non_terminal_nodes):        
        
        # Initialize lists that contain the nodes by color
        node_lists_by_color = []
        for color_index in range(r):
            utils.nullfunc(color_index)
            node_lists_by_color.append([])
        
        # Randomization: Color all words with the same stem one color randomly with r-1 colors
        #print "* Randomly coloring nodes..."
        for node_group in node_groups:
            rand_color = random.randint(0, r-1) # Color with r colors
            for node in node_group:

                # Do not color the start and end nodes
                if node == self.start_node or node == self.end_node:
                    continue

                # Set the node color
                node.set_color(rand_color)
                
                # Add this node to the node color list
                node_lists_by_color[rand_color].append(node)
                    
        # Create list of all colors
        all_colors = range(len(node_lists_by_color))

        # Clear node lists of different colors
        node_lists_of_different_color = []        
        for different_node_color in all_colors:
            node_lists_of_different_color.append([])
            for node_color in all_colors:
                if node_color == different_node_color:
                    continue
                node_lists_of_different_color[different_node_color].extend(node_lists_by_color[node_color])
        
        # Initialization: Define the colorful paths of length 0
        # Each non-start and non-end node is a colorful set
        # print "* Initializing nodes..."
        stopwords_len = len(self.stopwords)
        for v in all_non_terminal_nodes:
            edge_score = self.edge_trans[self.start_node.number][v.number][stopwords_len]
            v.add_color_set(\
                colorset.ColorSet(set([v.color]), edge_score,\
                                  [ config._NO_TRANS_INDEX ], [self.start_node, v], tuple([v.color])))

        # Length of stopwords list
        stopwords_len = len(self.stopwords)+1

        # Loop through path lengths from 0 to k-1
        for i in range(k):

            # Loop through all nodes
            for from_node in all_non_terminal_nodes:

                for to_node in node_lists_of_different_color[from_node.color]:

                    # Get all colorsets of size i for this node
                    for color_set_c in from_node.get_color_sets_by_size(i):
                        
                        # If the destination node color is not in this color set
                        if not to_node.color in color_set_c.colors:

                            # Explore all edges to the to node
                            for edge_stopword_index in range(stopwords_len):
                                edge_score = self.edge_trans[from_node.number][to_node.number][edge_stopword_index]
                                if edge_score == None:
                                    continue

                                # First generate the colorset key to see if this node already does not have this color set
                                new_color_set_colors = color_set_c.colors.copy()
                                new_color_set_colors.add(to_node.color)
                                new_color_set_key = colorset.get_color_set_key(new_color_set_colors)
                                new_color_set_score = self.__calc_iterative_score(color_set_c.score, edge_score)
                                
                                # If the this node does not have this color set, add it
                                if not to_node.has_color_set_by_colors(new_color_set_key) or \
                                   to_node.get_color_set_by_colors(new_color_set_key).score <= new_color_set_score:

                                    # Append edge list
                                    new_color_set_edge_list = color_set_c.edge_list[:]
                                    new_color_set_edge_list.append(edge_stopword_index)

                                    # Append node list
                                    new_color_set_node_list = color_set_c.node_list[:]
                                    new_color_set_node_list.append(to_node)

                                    new_color_set = colorset.ColorSet(new_color_set_colors, new_color_set_score,\
                                                                      new_color_set_edge_list, new_color_set_node_list,\
                                                                      new_color_set_key)
                                    to_node.add_color_set(new_color_set)
                                    
        # Return the top color set and score
        max_path_score = config._NEGATIVE_INFINITY
        max_node_list = None
        max_edge_list = None
        
        end_node_number = self.end_node.number
        for node in all_non_terminal_nodes:
            # Get all color sets of size k
            for color_set_k in node.get_color_sets_by_size(k):
                last_node_number = color_set_k.node_list[-1].number
                edge_score = self.edge_trans[last_node_number][end_node_number][-1]
                path_score = self.__calc_iterative_score(color_set_k.score, edge_score)

                if path_score > max_path_score:
                    max_path_score = path_score
                    max_node_list = color_set_k.node_list
                    max_edge_list = color_set_k.edge_list

        # Append edge to end node and end node to the max path
        max_edge_list.append(config._NO_TRANS_INDEX)
        max_node_list.append(self.end_node)

        print (max_edge_list, max_node_list, max_path_score)
        return (max_edge_list, max_node_list, max_path_score)

    def edge_iterator(self, from_node, to_node):

        # Skip the end node
        if from_node == self.end_node or to_node == self.start_node:
            return

        if from_node == self.start_node and to_node == self.end_node:
            return

        if from_node == to_node:
            return

        # Only direct links from start node
        if from_node == self.start_node:
            yield len(self.stopwords), self.edge_trans[from_node.number][to_node.number][-1]
            return

        # Only direct links to end node
        if to_node == self.end_node:
            yield len(self.stopwords), self.edge_trans[from_node.number][to_node.number][-1]
            return

        # Loop through all stop word transitions
        for stopword_index in range(len(self.stopwords)+1):
            score = self.edge_trans[from_node.number][to_node.number][stopword_index]
            if score != None:
                yield stopword_index, score
    
    def check_rep(self):
        print "* Checking representation..."
        
        all_nodes_except_start_end = set(self.nodes)
        all_nodes_except_start_end.remove(self.start_node)
        all_nodes_except_start_end.remove(self.end_node)        
        
        for node in self.nodes:
            if node == self.start_node:
                assert(set(node.out_edges.keys()) == all_nodes_except_start_end)
            elif node == self.end_node:
                assert(set(node.out_edges.keys()) == set())
            else:
                assert(len(node.out_edges.keys()) == len(all_nodes_except_start_end) - 1)
                        
    def __calc_iterative_score(self, phrase_score, edge_score):
        return phrase_score + edge_score

'''
    def make_ilp_file(self, max_path_length, filepath):
        # Trick so the graphs dont have to be regenerated
        # Renumber the nodes
        for i, node in enumerate(self.nodes):
            node.num = i
        outfile = open(filepath, 'w')

        # Formulate objective function
        outfile.write("max: ")
        for node in self.nodes:
            for out_edge in node.out_edges.values():
                outfile.write(str(math.log(out_edge.score))+" * i"+str(out_edge.from_node.num)+"_"+str(out_edge.to_node.num)+" + ")
        outfile.write("0;\n")

        # Formulate title length constraint
        for node in self.nodes:
            for out_edge in node.out_edges.values():
                outfile.write("i"+str(out_edge.from_node.num)+"_"+str(out_edge.to_node.num)+" + ")
        outfile.write("0 = "+str(max_path_length)+";\n")
        
        # Formulate only one start edge constraint
        for out_edge in self.start_node.out_edges.values():
            outfile.write("i"+str(self.start_node.num)+"_"+str(out_edge.to_node.num)+" + ")
        outfile.write("0 = 1;\n")

        # Formulate only one end edge constraint
        for node in self.nodes:
            if node.out_edges.has_key(self.end_node):
                outfile.write("i"+str(node.num)+"_"+str(self.end_node.num)+" + ")
        outfile.write("0 = 1;\n")
        
        # Formulate in degree = out degree constraint for each node except start and end node
        for node in self.nodes:
            # Omit start and end node
            if node == self.start_node or node == self.end_node:
                continue
            
            # Sum of the out edges
            for out_edge in node.out_edges.values():
                outfile.write("i"+str(node.num)+"_"+str(out_edge.to_node.num)+" + ")
            outfile.write("0 = ")
            # Sum of the in edges
            for in_node in self.nodes:
                if in_node.out_edges.has_key(node):
                    outfile.write("i"+str(in_node.num)+"_"+str(node.num)+" + ")
            outfile.write("0;\n")

        # Formulate out degree is at most one per stem group
        for node_group in self.node_groups_by_stem.values():
            for node in node_group:
                for out_edge in node.out_edges.values():
                    outfile.write("i"+str(node.num)+"_"+str(out_edge.to_node.num)+" + ")
            outfile.write("0 <= 1;\n")
                    
        # Formulate in degree is at most one per stem group
        for node_group in self.node_groups_by_stem.values():
            for node in node_group:
                for other_node in self.nodes:
                    if other_node.out_edges.has_key(node):
                         outfile.write("i"+str(other_node.num)+"_"+str(node.num)+" + ")
            outfile.write("0 <= 1;\n")
        
        # Formulate acyclic constraints
        # Flow in minus flow out = demand
        for node in self.nodes:
            # Sum of in flow
            for in_node in self.nodes:
                if in_node.out_edges.has_key(node):
                    outfile.write("f"+str(in_node.num)+"_"+str(node.num)+" + ")
            outfile.write("0 - ")
            
            # Sum of out flow
            for to_node in node.out_edges.keys():
                if node != to_node:
                    outfile.write("f"+str(node.num)+"_"+str(to_node.num)+" - ")
            outfile.write("0 = d"+str(node.num)+";\n")

        # Define demands
        # Demand is -max_path_length for the start node
        outfile.write("d"+str(self.start_node.num)+" = "+str(-1*max_path_length)+";\n")
        # Demand is 1 for all other nodes with an incoming edge and 0 otherwise
        for node in self.nodes:
            if node == self.start_node:
                continue
            outfile.write("d"+str(node.num)+" = ")
            for in_node in self.nodes:
                if in_node.out_edges.has_key(node):
                    outfile.write("i"+str(in_node.num)+"_"+str(node.num)+" + ")
            outfile.write("0 ;\n")

        # Capacity constraint: flow is less than edge capacity times the max path length
        for from_node in self.nodes:
            for to_node in from_node.out_edges.keys():
                outfile.write("f"+str(from_node.num)+"_"+str(to_node.num)+" <= "+str(max_path_length)\
                              +"* i"+str(from_node.num)+"_"+str(to_node.num)+";\n")
        
        # Formulate indicator variables
        varstr = ""
        for node in self.nodes:
            # Write capacitance variables
            if node != self.start_node:
                varstr += "int d"+str(node.num)+";\n"

            # Write edge variables
            for out_edge in node.out_edges.values():
                outfile.write("0 <= i"+str(out_edge.from_node.num)+"_"+str(out_edge.to_node.num)+" <= 1;\n")
                varstr += "int i"+str(out_edge.from_node.num)+"_"+str(out_edge.to_node.num)+";\n"
                varstr += "int f"+str(out_edge.from_node.num)+"_"+str(out_edge.to_node.num)+";\n"
        outfile.write(varstr)
        outfile.close()

    def ilpsearch(self, max_path_length, filepath):
        # Make ILP file
        self.make_ilp_file(max_path_length, filepath)

        print "* Solving..."
        # Run lpsolve
        t_timer = new_timer.Timer()
        lp_pipe = os.popen( config._LP_SOLVE + " -time "+filepath+" 2>&1")
        lp_output = lp_pipe.read()
        lp_pipe.close()
        total_time = t_timer.stotal()
        
        # Delete the generated file
        os.unlink(filepath)
        
        # Scan lpsolve output
        objective_value = config._NEGATIVE_INFINITY
        solving_time = "unknown"
        output_lines = []
        for line in lp_output.split("\n"):
            if line.startswith("Value of objective function:"):
                objective_value = float(line.split(": ", 1)[1])
                
            # Find indicator variables which are set to 1
            if line.endswith(" 1") and line.startswith("i"):
                output_lines.append(line)

            # Find the time for solving
            if line.startswith("CPU Time for solving:"):
                solving_time = line.split(" ")[4]
                    
        # ILP did not find a solution
        if len(output_lines) == 0:
            return tuple(["no_solution", config._NEGATIVE_INFINITY, total_time])

        # For each line which has an indicator variable set to 1
        # use it to rebuild the output
        ilp_word_nums = []
        for line in output_lines:
            [word_one_index, word_two_index] = line[1:].split(" ")[0].split("_")
            ilp_word_nums.append(tuple([int(word_one_index), int(word_two_index)]))

        # Start with the start node and go to the end node building the simple path found
        seek_word_num = self.start_node.num
        ilp_node_list = []
        while seek_word_num != self.end_node.num:
            for word_num_pair in ilp_word_nums:
                (first_word_num, second_word_num) = word_num_pair

                # If the first word is the one we are looking for
                if first_word_num == seek_word_num:
                    # Store this tuple in the list of nodes
                    ilp_node_list.append(self.nodes[first_word_num])
                    # Go to the next word
                    seek_word_num = second_word_num
                    # Remove this word pair
                    ilp_word_nums.remove(word_num_pair)
                    break

        # Add the end node to the node list
        ilp_node_list.append(self.end_node)
        
        # Make sure there are no extraneous edges (cycles)
        assert(len(ilp_word_nums) == 0)

        # Reconstruct the output phrase
        output_phrase = self.phrase_from_node_list(ilp_node_list)

        # This returns the time as calculated by python
        return tuple([output_phrase, objective_value, total_time])

        # This returns the time outputted by lp_solve
        # return tuple([output_phrase, objective_value, solving_time])
'''
