import lattice
import bigrams, config
import pickle, re, sys

def load_language_model(dataset_type):
    f = open(config._LANGUAGE_MODEL_PICKLE_PATH+dataset_type+".lang_model.pickle", 'r')
    lang_model = pickle.load(f)
    f.close()
    return lang_model

def save_language_model(dataset_type, lang_model):
    f = open(config._LANGUAGE_MODEL_PICKLE_PATH+dataset_type+".lang_model.pickle", 'w')
    pickle.dump(lang_model, f)
    f.close()

def triples(s):
    for i in range(len(s)-2):
        yield((s[i], s[i+1], s[i+2]))

class ThreePairs:

    def __init__(self, datafile, dataset):

        self.three_pairs = set()

        # Set the sentence splitter and split text into sentences
        if dataset == lattice._DATASET_ALGORITHMS or dataset == lattice._DATASET_REUTERS:
            bigrams.sentence_splitter = re.compile("#sentence#") # Algorithms or Reuters
        else:
            bigrams.sentence_splitter = re.compile("[\.|\?|\!|\;] ") # Physics or Civil

        # Get section titles and text
        (section_titles, section_text) = lattice.get_section_titles_and_text(datafile, dataset)

        for sec_num, text in enumerate(section_text):

            print "Analyzing for triples : "+str(sec_num)+" "+section_titles[sec_num]
            
            # Split into sentences
            sentences = bigrams.split_into_sentences(text)
            
            for sentence in sentences:
                
                # Split each sentence by punctuation
                punc_parts = bigrams.split_by_punctuation(sentence)
                
                for punc_part in punc_parts:

                    bigrams._STOP_WORD_FILTERING = True
                    bigrams._STEMMING = True

                    # Split into words without stop words
                    words = bigrams.split_into_words(punc_part)

                    # Add each three content word tuple
                    for three_word_tuple in triples(words):
                        self.three_pairs.add(three_word_tuple)

    def is_triple(self, word_one, word_two, word_three):
        if word_one == "#START#":
            return True
        elif word_three == "#END":
            return True
        else:
            return (word_one, word_two, word_three) in self.three_pairs

    def is_triple_tuple(self, word_triple):
        if word_triple[0] == "#START#" or word_triple[2] == "#END#":
            return True
        else:
            return word_triple in self.three_pairs
        
class LanguageModel:

    def __init__(self, data_file, dataset, cartesian_stopwords = [], single_stopwords = []):
        bigrams._STOP_WORD_FILTERING = False
        bigrams._STEMMING = False

        # Set the sentence splitter and split text into sentences
        if dataset == lattice._DATASET_ALGORITHMS or dataset == lattice._DATASET_REUTERS:
            bigrams.sentence_splitter = re.compile('#sentence#') # Algorithms or Reuters
        else:
            bigrams.sentence_splitter = re.compile('[\.|\?|\!|\;] ') # Physics or Civil
            
        self.dataset_type = dataset
        self.bg_counts = {}
        self.ug_counts = {}
        self.unigrams = set()
        
        # Add stopwords to unigrams set
        cartesian_stopwords_set = set(cartesian_stopwords)
        # Add stop word pairs
        self.unigrams = self.unigrams | cartesian_stopwords_set | set(single_stopwords)

        # Get section titles and text
        (section_titles, section_text) = lattice.get_section_titles_and_text(data_file, self.dataset_type)

        for sec_num, text in enumerate(section_text):

            print "Extracting bigrams: "+str(sec_num)+" "+section_titles[sec_num]
            
            # Split into sentences
            sentences = bigrams.split_into_sentences(text)
            
            for sentence in sentences:

                # Split each sentence by punctuation
                punc_parts = bigrams.split_by_punctuation(sentence)

                for punc_part in punc_parts:

                    # Get bigrams from this section
                    bigram_strings = bigrams.split_into_bigrams(punc_part)

                    # Update the unstemmed to stemmed count
                    for bigram_string in bigram_strings:
                        [word_one, word_two] = bigram_string.split(" ")
                        
                        # Update set of unigrams
                        self.unigrams.add(word_one)
                        self.unigrams.add(word_two)
                        
                        # Update bigram and unigram counts
                        self.bg_counts[(word_one, word_two)] =  self.bg_counts.get((word_one, word_two), 0)+1
                        self.ug_counts[word_one] = self.ug_counts.get(word_one, 0)+1
                        
                    # If a stop words list has been specified,
                    # extract all trigram pairs
                    if len(cartesian_stopwords) > 0:
                        
                        trigram_strings = bigrams.split_into_trigrams(punc_part)
                        
                        for trigram_string in trigram_strings:
                            [word_one, word_two, word_three] =  trigram_string.split(" ")
                            
                            # If the first two words in the trigam form are a stop word, increment the count
                            first_two_words = word_one+"_"+word_two
                            if first_two_words in cartesian_stopwords_set:
                                
                                self.unigrams.add(word_three)
                                
                                # Update bigram and unigram counts
                                self.bg_counts[(first_two_words, word_three)] =  self.bg_counts.get((first_two_words, word_three), 0)+1
                                self.ug_counts[first_two_words] = self.ug_counts.get(first_two_words, 0)+1

        sys.stdout.flush()
        
    def get(self, word_one, word_two):
        if config._ENABLE_LAPLACE_SMOOTHING:
            return 1.0 * (self.bg_counts.get((word_one, word_two), 0)+1) / (self.ug_counts.get(word_one, 0) + len(self.unigrams))
        else:
            return self.ug_counts.has_key(word_one) and ( 1.0 * self.bg_counts.get((word_one, word_two), 0) / self.ug_counts.get(word_one, 0) ) or 0.0

    def find_best_transition(self, word_one, word_two, stopwords):

        # First we need to find the transition that has
        # the maximum probability between these two
        # content words                    
        max_edge_score = 0.0
        max_edge_words = None

        # Calculate transition scores between this word
        # and the next word using stop words
        for stopword in stopwords:
            
            # Calculate transition probability for this stop word
            previous_word_prob = self.get(word_one, stopword)
            following_word_prob = self.get(stopword, word_two)
            prob = previous_word_prob * following_word_prob
                
            # Store this transition if it is the max
            if prob >= max_edge_score:
                max_edge_score = prob
                max_edge_words = stopword

                print str(prob)+word_one+" ["+stopword+"] "+word_two

        # Also consider the bigram score with no stop words
        prob = self.get(word_one, word_two)

        print str(prob)+word_one+" ["+""+"] "+word_two        
        # Store this transition if it is the max
        if prob >= max_edge_score:
            max_edge_score = prob
            max_edge_words = ""

        return (max_edge_words, max_edge_score)
