package main;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Set;

import utilities.Utils;
import entities.Section;


public class FeatureVectorBuilder {

	public static void BuildFeatureVectors(List<Section> sections, String trainoutfilepath, String testoutfilepath, String testwordsoutfilepath, double percentTraining, boolean duplicationRemoval) {
		
		System.out.print("Initializing Feature Vector Builder...");
		
		// Calculate the number of sections
		int numberOfSections = sections.size();
		Hashtable<String, Integer> numberOfSectionsWithWord = new Hashtable<String, Integer>();
		Hashtable<String, Integer> firstSectionWithWord = new Hashtable<String, Integer>();
		Hashtable<String, Integer> firstSentenceWithWord = new Hashtable<String, Integer>();
		Hashtable<String, Integer> allWordCount = new Hashtable<String, Integer>();
		
		// Build a Set of all words in the section bodies
		Set<String> allWords = new HashSet<String>();
		for(Section section : sections) {
			List<List<String>> sectionSentences = section.getSentences();
			for(List<String> sentence : sectionSentences) {
				for(String word : sentence) {
					
					allWords.add(word);
					
					if(!numberOfSectionsWithWord.containsKey(word)) {
						numberOfSectionsWithWord.put(word, 0);
					}
					
					if(!firstSectionWithWord.containsKey(word)) {
						firstSectionWithWord.put(word, -1);
					}
					
					if(!firstSentenceWithWord.containsKey(word)) {
						firstSentenceWithWord.put(word, -1);
					}
					
					if(!allWordCount.containsKey(word)) {
						allWordCount.put(word, 0);
					}
					allWordCount.put(word, allWordCount.get(word)+1);
					
				}
			}
		}
		
		for(int sectionNum = 0; sectionNum < sections.size(); sectionNum++) {
			Section section = sections.get(sectionNum);
			// Calculate the number of sections each word in the appears
			for(String word : allWords) {
				
				if(section.bodyContainsWord(word)) {
					numberOfSectionsWithWord.put(word, numberOfSectionsWithWord.get(word)+1);
				
					// Calculate the section in which the word first appears
					if(firstSectionWithWord.get(word) == -1) {
						firstSectionWithWord.put(word, sectionNum+1);
					}
				}				
			}
		}

		try {
			boolean isTrainingFile = true;
			BufferedWriter out = new BufferedWriter(new FileWriter(trainoutfilepath));
			BufferedWriter wordsout = new BufferedWriter(new FileWriter(testwordsoutfilepath));
			
			int i=1;
			int trainingSections = 0;
			int numberOfFeatureVectors = 0;
			// Generate feature vectors
			for(Section section : sections) {
				Set<String> sectionWordsAnalyzed = new HashSet<String>();
				
				// Switch to test file output
				if(isTrainingFile && 1.0*i/sections.size() > percentTraining) {
					isTrainingFile = false;
					out.close();
					out = new BufferedWriter(new FileWriter(testoutfilepath));
				}
				
				// Count the number of sections used for training
				if(isTrainingFile) {
					trainingSections++;
				}
				
				System.out.println(i++ +" of "+sections.size()+" "+section.getTitleString());
				for(List<String> sentence : section.getSentences()) {
					for(String word : sentence) {
						
						if(duplicationRemoval == true) {
							// Only analyze unique words
							if(!sectionWordsAnalyzed.contains(word)) {
								sectionWordsAnalyzed.add(word);
							} else {
								continue;
							}
						}
							
						String featureIsInTitle = section.isWordInTitle(word)?"+title":"-title";
						double featureFirstSectionWithWord = 1.0 * firstSectionWithWord.get(word)/numberOfSections;
						double featureNumberOfSectionsWithWord = 1.0 * numberOfSectionsWithWord.get(word)/numberOfSections;
						double featureRelativeFrequency = 1.0 * section.getWordOccurences(word)/allWordCount.get(word); // TODO
						double featureFirstWord = 1.0 * section.getFirstWordIndex(word)/section.getNumberOfTotalBodyWords();  // TODO
						double featureFirstSentenceWithWord = 1.0*section.getFirstSentenceIndexWithWord(word)/section.getSentences().size();
						int featureRankOfWordInSection = section.getFrequencyRank(word);
						double featureTDFIDF = (1.0 * section.getWordOccurences(word)/section.getNumberOfTotalBodyWords()) 
													* Math.log(1.0 * numberOfSections/numberOfSectionsWithWord.get(word));
						
						numberOfFeatureVectors++;
						
						// Write test words to file
						if(isTrainingFile == false) {
							wordsout.write(word+Utils.END_OF_LINE);
						}
						
						out.write(
								featureFirstSectionWithWord+", "+
								featureNumberOfSectionsWithWord+", "+
								featureRelativeFrequency+", "+
								featureFirstWord+", "+
								featureFirstSentenceWithWord+", "+
								featureRankOfWordInSection+", "+
								featureTDFIDF+", "+
								featureIsInTitle+". "+Utils.END_OF_LINE);
					}
				}
			}
			out.close();
			wordsout.close();
			
			// Display Stastics
			System.out.println("Unique Words: "+allWords.size());
			System.out.println("Avg. Unique Words Per Section: "+1.0*allWords.size()/numberOfSections);
			System.out.println("Total Feature Vectors: "+numberOfFeatureVectors);
			System.out.println("Average Feature Vectors Per Section: "+1.0*numberOfFeatureVectors/numberOfSections);
			System.out.println("Training Sections: "+trainingSections);
			System.out.println("Test Sections: "+(sections.size()-trainingSections));
		} catch (IOException e) {
			System.out.println(" ** Error: "+e.toString());
		}
	}

}
