package main;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Set;

import utilities.Utils;
import entities.SGram;
import entities.Section;
import entities.Tagger;

public class HierarchicalFeatureVectorBuilder {
	public static final boolean FEATURE_FIRST_SECTION_WITH_WORD = true;
	public static final boolean FEATURE_NUMBER_OF_SECTIONS_WITH_WORD = true;
	public static final boolean FEATURE_RELATIVE_FREQUENCY = true;
	public static final boolean FEATURE_FIRST_WORD = true;
	public static final boolean FEATURE_FIRST_SENTENCE_WITH_WORD = true;
	public static final boolean FEATURE_RANK_OF_WORD_IN_SECTION = true;
	public static final boolean FEATURE_TFIDF = true;
	public static final boolean FEATURE_IS_NUMBER = false;
	public static final boolean FEATURE_IN_PARENT_TITLE = false;
	public static final boolean FEATURE_IN_GRANDPARENT_TITLE = false;
	public static final boolean FEATURE_IN_SIBLING_TITLE = false;
	public static final boolean FEATURE_IN_SIBLING_BODY = false;
	public static final boolean FEATURE_IN_PARENT_BODY = false;
	public static final boolean FEATURE_POS_TAG = false;
	public static final boolean FEATURE_SGRAMS = false;
	public static final boolean FEATURE_IS_FIRST_SENTENCE = false;
	
	private static final boolean SKIP_IGNORE = false;
	
	private static final String PREFIX_TRAIN = "#TRAIN";
	private static final String PREFIX_TEST = "#TEST"; 
	
	public static String GenerateFlatFile(List<Section> sections) {
		String generated = "";
		for(Section section : sections) {
			List<List<String>> sectionSentences = section.getSentences();
			if(sectionSentences.size() > 0) {
				System.out.print("section "+section.getTitleString()+"\n");
				for(List<String> sentence : sectionSentences) {
					for(String word : sentence) {
						System.out.print(word+" ");
					}
					System.out.print("---\n");
				}
			}
		}
		return generated;
	}
	
	public static void BuildFeatureVectors(List<Section> sections, 
			String trainoutfilepath, String testoutfilepath,
			String testwordsoutfilepath, String sectionnamesoutfilepath, 
			String featurenamesoutfilepath, boolean includeHierarchicalFeatures,
			boolean includeLexicalFeatures,
			double percentTraining,
			Tagger tagger,
			List<SGram> sgrams) {
		
		System.out.print("Initializing Feature Vector Builder...");
		
		// Calculate the number of sections
		int numberOfSections = sections.size();
		Hashtable<String, Integer> numberOfSectionsWithWord = new Hashtable<String, Integer>();
		Hashtable<String, Integer> firstSectionWithWord = new Hashtable<String, Integer>();
		Hashtable<String, Integer> firstSentenceWithWord = new Hashtable<String, Integer>();
		Hashtable<String, Integer> allWordCount = new Hashtable<String, Integer>();
		
		// Build a Set of all words in the section bodies
		Set<String> allWords = new HashSet<String>();
		for(Section section : sections) {
			List<List<String>> sectionSentences = section.getSentences();
			for(List<String> sentence : sectionSentences) {
				for(String word : sentence) {
					
					allWords.add(word);
					
					if(!numberOfSectionsWithWord.containsKey(word)) {
						numberOfSectionsWithWord.put(word, 0);
					}
					
					if(!firstSectionWithWord.containsKey(word)) {
						firstSectionWithWord.put(word, -1);
					}
					
					if(!firstSentenceWithWord.containsKey(word)) {
						firstSentenceWithWord.put(word, -1);
					}
					
					if(!allWordCount.containsKey(word)) {
						allWordCount.put(word, 0);
					}
					allWordCount.put(word, allWordCount.get(word)+1);
					
				}
			}
		}
		
		for(int sectionNum = 0; sectionNum < sections.size(); sectionNum++) {
			Section section = sections.get(sectionNum);
			// Calculate the number of sections each word in the appears
			for(String word : allWords) {
				
				if(section.bodyContainsWord(word)) {
					numberOfSectionsWithWord.put(word, numberOfSectionsWithWord.get(word)+1);
				
					// Calculate the section in which the word first appears
					if(firstSectionWithWord.get(word) == -1) {
						firstSectionWithWord.put(word, sectionNum+1);
					}
				}
			}
		}

		try {
			boolean isTrainingFile = true;
			BufferedWriter out = new BufferedWriter(new FileWriter(trainoutfilepath));
			BufferedWriter wordsout = new BufferedWriter(new FileWriter(testwordsoutfilepath));
			BufferedWriter sectionsout = new BufferedWriter(new FileWriter(sectionnamesoutfilepath));
			
			int i=1;
			int trainingSections = 0;
			int numberOfFeatureVectors = 0;
			
			// Generate feature vectors
			for(int sectionIndex = 0; sectionIndex < sections.size(); sectionIndex++) {

				// Switch to test file output
				if(isTrainingFile && 1.0*i/sections.size() > percentTraining) {
					isTrainingFile = false;
					out.close();
					out = new BufferedWriter(new FileWriter(testoutfilepath));
				}

				// Get the section
				Section section = sections.get(sectionIndex);
				sectionsout.write((isTrainingFile?PREFIX_TRAIN:PREFIX_TEST)+" "+section.getTitleString()+Utils.END_OF_LINE);
								
				// Count the number of sections used for training
				if(isTrainingFile) {
					trainingSections++;
				}
				
				// If this section is supposed to be ignored, skip generating feature vectors for it
				if(SKIP_IGNORE && section.isIgnore()) {
					System.out.println("SKIPPING SECTION: "+section.getTitleString());
					continue;
				}
				
				// Initialize set of words analyzed in this section
				Set<String> sectionWordsAnalyzed = new HashSet<String>();
				
				
				System.out.println(i++ +" of "+sections.size()+" "+section.getTitleString()+" Unique Words: "+section.getNumberOfTotalUniqueBodyWords()+" Sentences: "+section.getSentences().size());
				int sentenceNumber = 0;
				
				for(List<String> sentence : section.getSentences()) {
					
					// Increment the sentence number
					sentenceNumber++;
					
					// Per sentence sgram features
					String featureSGrams = "";
					if(FEATURE_SGRAMS) {
						for(SGram sgram : sgrams) {
							featureSGrams += ( (sgram.sentenceHasSgram(sentence)) ? "1" : "0" ) + ", ";
						}
					}
					
					for(String word : sentence) {
						
						// Only analyze unique words
						if(!sectionWordsAnalyzed.contains(word)) {
							sectionWordsAnalyzed.add(word);
						} else {
							continue;
						}
							
						String isInTitle = section.isWordInTitle(word)?"+title":"-title";
						double featureFirstSectionWithWord = 1.0 * firstSectionWithWord.get(word)/numberOfSections;
						double featureNumberOfSectionsWithWord = 1.0 * numberOfSectionsWithWord.get(word)/numberOfSections;
						double featureRelativeFrequency = 1.0 * (1.0 * section.getWordOccurences(word)/section.getNumberOfTotalBodyWords()) / allWordCount.get(word); // TODO
						double featureFirstWord = 1.0 * section.getFirstWordIndex(word)/section.getNumberOfTotalBodyWords();  // TODO
						double featureFirstSentenceWithWord = 1.0*section.getFirstSentenceIndexWithWord(word)/section.getSentences().size();
						int featureRankOfWordInSection = section.getFrequencyRank(word);
						double featureTDFIDF = (1.0 * section.getWordOccurences(word)/section.getNumberOfTotalBodyWords()) 
													* Math.log(1.0 * numberOfSections/numberOfSectionsWithWord.get(word));
						String featureIsNumber = Utils.isNumeric(word)?"1":"0";
						String featureWordInParentTitle = (section.getParent()!=null && section.getParent().isWordInTitle(word))?"1":"0";
						String featureWordInGrandParentTitle = (section.getParent()!=null && section.getParent().getParent()!=null && section.getParent().getParent().isWordInTitle(word))?"1":"0";
						String featureWordInAnySiblingTitle = section.wordInAnySiblingTitle(word)?"1":"0";
						String featureWordInAnySiblingBody = section.wordInAnySiblingBody(word)?"1":"0";
						String featureWordInParentBody = section.wordInParentBody(word)?"1":"0";
						String featurePOSTag = (FEATURE_POS_TAG)?tagger.get(sectionIndex, word):"";
						String featureFirstSentence = section.getDepth()+"";
						numberOfFeatureVectors++;
						
						// If we are generating a test file, write the words to file
						if(!isTrainingFile) { wordsout.write(word+" "+sectionIndex+Utils.END_OF_LINE); }

						out.write(
								(FEATURE_FIRST_SECTION_WITH_WORD?(featureFirstSectionWithWord+", "):"")+
								(FEATURE_NUMBER_OF_SECTIONS_WITH_WORD?(featureNumberOfSectionsWithWord+", "):"")+
								(FEATURE_RELATIVE_FREQUENCY?(featureRelativeFrequency+", "):"")+
								(FEATURE_FIRST_WORD?(featureFirstWord+", "):"")+
								(FEATURE_FIRST_SENTENCE_WITH_WORD?(featureFirstSentenceWithWord+", "):"")+
								(FEATURE_RANK_OF_WORD_IN_SECTION?(featureRankOfWordInSection+", "):"")+
								(FEATURE_TFIDF?(featureTDFIDF+", "):"")+
								(FEATURE_IS_NUMBER?(featureIsNumber+", "):"")+
								(FEATURE_POS_TAG?(featurePOSTag+", "):"")+
								(FEATURE_IS_FIRST_SENTENCE?(featureFirstSentence+", "):"")+
								((includeHierarchicalFeatures && FEATURE_IN_PARENT_TITLE)?(featureWordInParentTitle+", "):"")+
								((includeHierarchicalFeatures && FEATURE_IN_GRANDPARENT_TITLE)?(featureWordInGrandParentTitle+", "):"")+
								((includeHierarchicalFeatures && FEATURE_IN_SIBLING_TITLE)?(featureWordInAnySiblingTitle+", "):"")+
								((includeHierarchicalFeatures && FEATURE_IN_SIBLING_BODY)?(featureWordInAnySiblingBody+", "):"")+
								((includeHierarchicalFeatures && FEATURE_IN_PARENT_BODY)?(featureWordInParentBody+", "):"")+
								(includeLexicalFeatures?(word+", "):"")+
								(FEATURE_SGRAMS?featureSGrams:"")+
								isInTitle+". "+Utils.END_OF_LINE);
					}
				}
			}
			sectionsout.close();
			out.close();
			wordsout.close();
			
			// Write Boostexter names file
			BufferedWriter featurenamesout = new BufferedWriter(new FileWriter(featurenamesoutfilepath));
			featurenamesout.write("+title, -title."+Utils.END_OF_LINE);
			if(FEATURE_FIRST_SECTION_WITH_WORD)featurenamesout.write("featureFirstSectionWithWord: continuous."+Utils.END_OF_LINE);
			if(FEATURE_NUMBER_OF_SECTIONS_WITH_WORD)featurenamesout.write("featureNumberOfSectionsWithWord: continuous."+Utils.END_OF_LINE);
			if(FEATURE_RELATIVE_FREQUENCY)featurenamesout.write("featureRelativeFrequency: continuous."+Utils.END_OF_LINE);
			if(FEATURE_FIRST_WORD)featurenamesout.write("featureFirstWord: continuous."+Utils.END_OF_LINE);
			if(FEATURE_FIRST_SENTENCE_WITH_WORD)featurenamesout.write("featureFirstSentenceWithWord: continuous."+Utils.END_OF_LINE);
			if(FEATURE_RANK_OF_WORD_IN_SECTION)featurenamesout.write("featureRankOfWordInSection: continuous."+Utils.END_OF_LINE);
			if(FEATURE_TFIDF)featurenamesout.write("featureTDFIDF: continuous."+Utils.END_OF_LINE);
			if(FEATURE_IS_NUMBER)featurenamesout.write("featureIsNumber: continuous."+Utils.END_OF_LINE);
			if(FEATURE_POS_TAG)featurenamesout.write("featurePOStags: "+Utils.join(tagger.allPOSTags, ", ")+"."+Utils.END_OF_LINE);
			if(FEATURE_IS_FIRST_SENTENCE)featurenamesout.write("featureIsFirstSentence: 0, 1, 2, 3, 4, 5, 6, 7."+Utils.END_OF_LINE);
			if(includeHierarchicalFeatures && FEATURE_IN_PARENT_TITLE)featurenamesout.write("featureWordInParentTitle: continuous."+Utils.END_OF_LINE);
			if(includeHierarchicalFeatures && FEATURE_IN_GRANDPARENT_TITLE)featurenamesout.write("featureWordInGrandParentTitle: continuous."+Utils.END_OF_LINE);
			if(includeHierarchicalFeatures && FEATURE_IN_SIBLING_TITLE)featurenamesout.write("featureWordInAnySiblingTitle: continuous."+Utils.END_OF_LINE);
			if(includeHierarchicalFeatures && FEATURE_IN_SIBLING_BODY)featurenamesout.write("featureWordInAnySiblingBody: continuous."+Utils.END_OF_LINE);
			if(includeHierarchicalFeatures && FEATURE_IN_PARENT_BODY)featurenamesout.write("featureWordInParentBody: continuous."+Utils.END_OF_LINE);
			if(includeLexicalFeatures) {
				featurenamesout.write("featureStem: ");
				int wordCounter = 0;
				for(String word : allWords) {
					wordCounter++;
					featurenamesout.write(word+((wordCounter!=allWords.size())?", ":""));
				}
				featurenamesout.write("."+Utils.END_OF_LINE);
			}
			if(FEATURE_SGRAMS) {
				for(int sgramNumber = 0; sgramNumber < sgrams.size(); sgramNumber++) {
					featurenamesout.write("featureSgram"+sgramNumber+": 0, 1."+Utils.END_OF_LINE);
				}
			}
			featurenamesout.close();
			
			// Display Stastics
			System.out.println("<li>Unique Words: "+allWords.size()+"</li>");
			System.out.println("<li>Avg. Unique Words Per Section: "+1.0*allWords.size()/numberOfSections+"</li>");
			System.out.println("<li>Total Feature Vectors: "+numberOfFeatureVectors+"</li>");
			System.out.println("<li>Average Feature Vectors Per Section: "+1.0*numberOfFeatureVectors/numberOfSections+"</li>");
			System.out.println("<li>Training Sections: "+trainingSections+"</li>");
			System.out.println("<li>Test Sections: "+(sections.size()-trainingSections)+"</li>");
			
		} catch (IOException e) {
			System.out.println(" ** Error: "+e.toString());
		}
	}

}
