#*****************************************************************************
#       Copyright (C) 2001 Fernando Prez. <fperez@colorado.edu>
#
#  Distributed under the terms of the GNU Lesser General Public License (LGPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#  The full text of the LGPL is available at:
#
#                  http://www.gnu.org/copyleft/lesser.html
#*****************************************************************************

__author__ = 'Fernando Prez. <fperez@colorado.edu>'
__license__ = 'LGPL'
__version__= '0.1'
__date__   = 'Thu Apr 10 10:48:44 MDT 2003'

import resource

class Timer:
    """Stopwatch-like class to take timings during code execution.

    It is not meant to be an ultra-accurate tool, since it doesn't calibrate
    its own overhead.  But it will work well for timing reasonably long
    running codes.

    It is designed to be used as a split-timing stopwatch, with which one can
    collect timing information at various checkpoints of a long running
    program."""

    def __init__(self):
        """Return a new timer."""
        # Keep track of user times
        self.start,self.sstart = self.clocks()
        self.times = [self.start]
        self.splits = [0]
        self.tagged_splits = []
        # And system ones
        self.stimes = [self.sstart]
        self.ssplits = [0]
        self.tagged_ssplits = []

    def clock(self):
        """A clock function without the wraparound problems of time.clock()

        Returns the user time for the current process via getrusage()."""
        
        return resource.getrusage(resource.RUSAGE_SELF)[0]

    def sclock(self):
        """Similar to clock(), but return system time instead."""
        
        return resource.getrusage(resource.RUSAGE_SELF)[1]

    def clocks(self):
        """Return both user and system time."""
        
        return resource.getrusage(resource.RUSAGE_SELF)[0:2]

    def split(self,tag=None):
        """Take a split time reading."""
        ut,st = self.clocks()
        self.times.append(ut)
        self.stimes.append(st)
        self.splits.append(self.times[-1]-self.times[-2])
        self.ssplits.append(self.stimes[-1]-self.stimes[-2])
        if tag is not None:
            self.tagged_splits.append((tag,self.splits[-1],self.ssplits[-1]))

    def split_print(self,tag=None):
        """Take a split and print it."""

        self.split(tag)
        if tag is None:
            print "Last split timing (user/sys): %s / %s s." % \
                  (self.splits[-1],self.ssplits[-1])
        else:
            print "Split timing for <%s> (user/sys): %s / %s s." % \
                  (tag,self.splits[-1],self.ssplits[-1])

    def print_tagged_splits(self):
        """Print a formatted table of all tagged splits."""
        print 'Summary of tagged splits (user/sys)'
        print '-----------------------------------'
        for tsplit in self.tagged_splits:
            print '%s: %s / %s s' % tsplit
            
    def last_split(self):
        """Return the last user split."""

        return self.splits[-1]

    def last_ssplit(self):
        """Return the last system split."""

        return self.ssplits[-1]

    def last_splits(self):
        """Return the last user and system split."""

        return (self.splits[-1],self.ssplits[-1])

    def total(self):
        """Return total user time elapsed since timer's creation."""
        return self.clock() - self.start

    def stotal(self):
        """Return total system time elapsed since timer's creation."""
        return self.sclock() - self.sstart

    def totals(self):
        """Return total user and system time elapsed since timer's
        creation."""
        unow,snow = self.clocks()
        return (unow - self.start,snow-self.sstart)
