import find_optimal_colors, bigrams, graph, language_model, utils, lattice, tagger, terminal_word_probs, config, srilm_arpa, srilm_tagger, new_timer, use_psyco
import sys, os, getopt, math

use_psyco.nullfunc()

_BINARY_SELECTION_SCORES = False
_INFINITY = 100000000
_SHREDDER_PATH = "http://shredder.csail.mit.edu:8010/"

def get_graph_pickle_path(path, section_number, job_id):
    return path+str(section_number)+'.pickle'+'_'+job_id

def load_ilp_scores(ilp_output_file):
    f = open(ilp_output_file, 'r')
    lines = f.readlines()
    f.close()

    ilp_scores_dict = {}
    section_number = ""
    for line in lines:
        line = line.strip()
        
        # Get section number
        if line.startswith("Section"):
            [section_string, hash_string, section_number_string, actual_title] = line.split(" ",3)
            section_number = section_number_string
            
        # Get ILP score
        elif line.startswith("ILP:"):
            [ignore, score] = line.rsplit(" ", 1)
            ilp_scores_dict[section_number] = score

    return ilp_scores_dict

def extract_ilp_score_from_ilp_output(ilp_outfile):
    f = open(ilp_outfile)
    line = f.read().strip()
    f.close()

    [method, phrase, time, score] = line.split("---")
    
    return float(score)

def compare_to_ilp_score(rand_score, ilp_score):
    return float('%.3f'%math.log(float(rand_score))) >= float('%.3f'%math.log(float(ilp_score)))

def set_lock(lockpath):
    f = open(lockpath, 'w')
    f.write("locked")
    f.close()
    return

def remove_lock(lockpath):
    os.unlink(lockpath)
    return

def check_lock(lockpath):
    try:
        # If graph file exists, assume graph has been constructed
        f = open(lockpath, 'r')
        f.close()
        return True
        
    except:
        return False

def extract_ilp_score_from_lp_solve_output(lp_solve_output_filepath):
    ilp_score = 0.0
    
    # If this is an lp_solve solution file
    # Check if it has been solved
    f = open(lp_solve_output_filepath, 'r')
    lines = f.readlines()
    f.close()
    
    for line in lines:
        # Get score
        if "of objective" in line:
            ilp_score = float(line.strip().split(" ")[4])
            break

    # Make sure an ilp_score was extracted
    assert(ilp_score != 0.0)

    return ilp_score
        
def run_rand_iters(g, iters, number_of_content_words, outfile, ilp_score = 0.0):

    # Run Randomized Search
    top_rand_score = config._NEGATIVE_INFINITY
    top_rand_phrase = ""
    total_time = 0.0
    
    # Calculate optimal number of colors
    number_colors = find_optimal_colors.find_optimal_colors(number_of_content_words)
    node_groups = g.node_groups_by_stem.values()
    non_terminal_nodes = g.nodes[:]
    non_terminal_nodes.remove(g.start_node)
    non_terminal_nodes.remove(g.end_node)
        
    print "* Running Randomized Search for "+str(iters)+" iterations..."
    for iter_num in range(iters):
        print "* Running iteration "+str(iter_num)+"..."

        # Decolor graph
        print "* Decoloring graph..."
        g.decolor()

        # Run randomized algorithm
        print "* Running randomized search..."
        t_timer = new_timer.Timer()
        [rand_edges, rand_nodes, rand_score] = g.randomized_search(number_of_content_words, number_colors, node_groups, non_terminal_nodes)
        iter_time = t_timer.total()

        # Reconstruct phrase
        rand_phrase = g.phrase_from_edge_and_node_list(rand_edges, rand_nodes)
        
        # Update time taken
        total_time += iter_time

        # Write results to file
        print "RAND---"+str(rand_phrase)+"---"+str(iter_time)+"s---"+str(rand_score)
        utils.append_to_file_newline(outfile, "RAND---"+str(rand_phrase)+"---"+str(iter_time)+"s---"+str(rand_score))

        if compare_to_ilp_score(rand_score, ilp_score):
            print "ILP SCORE MATCHED."
            break
            
    print "TOP RAND---"+str(top_rand_phrase)+"---"+str(1.0 * total_time/iters)+"s---"+str(top_rand_score)
    utils.append_to_file_newline(outfile, "TOP RAND---"+str(top_rand_phrase)+"---"+str(1.0 * total_time/iters)+"s---"+str(top_rand_score))

'''
def search_ilp(section_number, section_title, dataset_type, graph_path, formulation_dir, job_id):
    # Convert section number to string
    section_number = str(section_number)

    # Set ILP output path
    ilp_output_path = config._OUTPUT_PATH + dataset_type + section_number + ".ilpout_"+job_id
    
    # Check if already solved
    if check_lock(ilp_output_path):
        print "Already solved, skipping..."
        return
            
    # Load or generate graph
    print "* Loading graph for Section # "+section_number+" "+section_title+"..."
    sys.stdout.flush()
    
    # Set file lock path
    file_lock_path = config._ILP_LOCK_DIR + dataset_type + "_" + section_number + "_ilpsearch.lockfile_" + job_id
    
    # If the graph is locked skip
    if check_lock(file_lock_path):
        print "File locked, skipping..."
        return
        
    # Set file lock so that no other processes attempt to build this graph 
    print "Setting file lock..."
    set_lock(file_lock_path)
    sys.stdout.flush()                
        
    # Build graph
    print "* Loading graph..."
    try:
        g = utils.load_pickle(get_graph_pickle_path(graph_path, section_number, job_id))
    except:
        print "* Loading Failed skipping..."
        return
    sys.stdout.flush()
    
    # Solving with ILP
    # Get number of content words in the title
    bigrams._STOP_WORD_FILTERING = True
    number_of_content_words = len(bigrams.split_into_words(section_title.lower()))
    # number_of_content_words = 3
    
    # Generate the lpsolve file
    print "* Generating lp_solve file..."
    lpfilename = formulation_dir + dataset_type + "_" + str(section_number) + ".lpfile_"+job_id
    (ilp_phrase, ilp_score, ilp_time) = g.ilpsearch(number_of_content_words+1, lpfilename)
    print "ILP: "+ilp_phrase+" "+str(math.e**float(ilp_score))+" "+str(ilp_time)
    f = open(ilp_output_path, 'w')
    f.write("ILP---"+ilp_phrase+"---"+str(ilp_time)+"---"+str(math.e**float(ilp_score)))
    f.close()
    
    # Remove building file lock
    print "* Removing file lock..."
    remove_lock(file_lock_path)
    sys.stdout.flush()

    return
'''
def search_beam_loop(section_titles, dataset_type, beam_size, graph_path, maxent_stub, job_id):

    stemmed_section_word_scores = {}
    print "* Loading binary selection scores..."
    binary_stemmed_section_word_scores = lattice.load_binary_selection_scores(maxent_stub+".sectiontitles", maxent_stub+".testwords")
    
    print "* Loading section words stemmed with scores..."
    non_binary_stemmed_section_word_scores = lattice.get_stemmed_section_word_scores(maxent_stub+".maxent_out", maxent_stub+".testwords")
    
    if _BINARY_SELECTION_SCORES:
        stemmed_section_word_scores = binary_stemmed_section_word_scores
    else:
        stemmed_section_word_scores = non_binary_stemmed_section_word_scores
  
    section_numbers = sorted(stemmed_section_word_scores.keys())

    for section_number in section_numbers:
        search_beam(section_number, section_titles[ str( section_number ) ], dataset_type, beam_size, graph_path, job_id)

def search_beam(section_number, section_title, dataset_type, beam_size, graph_path, job_id):

    # Convert section number to string
    section_number = str(section_number)    
    
    beam_output_path = config._OUTPUT_PATH + dataset_type + section_number + ".beamout_"+str(beam_size)+"_"+job_id
    
    # Set file lock path
    file_lock_path = config._ILP_LOCK_DIR+dataset_type+"_"+section_number+"_beam_"+str(beam_size)+".lockfile"+"_"+job_id
    
    # If the graph is locked skip
    if check_lock(file_lock_path):
        print "File locked, skipping..."
        return
    
    # Check if already computed
    if check_lock(beam_output_path):
        print beam_output_path
        print "Already computed, skipping..."
        return

    # Set file lock so that no other processes attempt to build this graph 
    print "Setting file lock..."
    set_lock(file_lock_path)
    sys.stdout.flush()                

    # Load graph
    print "* Loading graph for Section # "+str(section_number)+" "+section_title+"..."
    sys.stdout.flush()
    try:
        print "* Loading from AFS..."
        g = utils.load_pickle(get_graph_pickle_path(graph_path, section_number, job_id))
        
    except:

        print "* Loading from web..."
        try:
            g = utils.load_web_pickle(get_graph_pickle_path(_SHREDDER_PATH, section_number, job_id))
        except:
            print "* Loading Failed skipping..."
        
            # Remove building file lock
            print "* Removing file lock..."
            remove_lock(file_lock_path)
            sys.stdout.flush()
            return
    
    sys.stdout.flush()
            
    # Number of content words that are both in the title and section text
    # number_of_content_words = int(sum(binary_stemmed_section_word_scores[section_number].values()))
    # Get number of content words in the title
    bigrams._STOP_WORD_FILTERING = True
    number_of_content_words = len(bigrams.split_into_words(section_title.lower()))
    # number_of_content_words = 3
    
    # No file locking
    # Run Beam Search
    print "* Running Beam Search with size "+str(beam_size)+"..."
    [beam_score, beam_edges, beam_nodes, beam_time] = g.beam_search(number_of_content_words, beam_size)
    
    # Reconstruct phrase
    beam_phrase = g.phrase_from_edge_and_node_list(beam_edges, beam_nodes)
    
    print "BEAM---"+beam_phrase+"---"+str(beam_time)+"s---"+str(beam_score)
    f = open(beam_output_path, 'w')
    f.write("BEAM---"+beam_phrase+"---"+str(beam_time)+"s---"+str(beam_score))
    f.close()
    
    # Remove building file lock
    print "* Removing file lock..."
    remove_lock(file_lock_path)
    sys.stdout.flush()

    return

def search_rand_loop( section_titles, num_max_iters, dataset_type, graph_path, ilp_outfile, maxent_stub, job_id):

    stemmed_section_word_scores = {}
    print "* Loading binary selection scores..."
    binary_stemmed_section_word_scores = lattice.load_binary_selection_scores(maxent_stub+".sectiontitles", maxent_stub+".testwords")
    
    print "* Loading section words stemmed with scores..."
    non_binary_stemmed_section_word_scores = lattice.get_stemmed_section_word_scores(maxent_stub+".maxent_out", maxent_stub+".testwords")

    if _BINARY_SELECTION_SCORES:
        stemmed_section_word_scores = binary_stemmed_section_word_scores
    else:
        stemmed_section_word_scores = non_binary_stemmed_section_word_scores
  
    section_numbers = sorted(stemmed_section_word_scores.keys())

    for section_number in section_numbers:
        search_rand(section_number, section_titles[ str( section_number ) ] , num_max_iters, dataset_type, graph_path, ilp_outfile, job_id)

def search_rand(section_number, section_title, num_max_iters, dataset_type, graph_path, ilp_outfile, job_id):
        
    # Get ILP Score
    # If no ilp output file specified, run forever
    if ilp_outfile == None or ilp_outfile == "":
        ilp_score = _INFINITY
    else:
        # Use this to read from lpsolve output file
        # ilp_score = extract_ilp_score_from_lp_solve_output(ilp_outfile)

        # Use this to read from ilp output file
        ilp_score = extract_ilp_score_from_ilp_output(ilp_outfile)
        
    # Set output path
    rand_output_path = config._OUTPUT_PATH + dataset_type + str(section_number) + ".randout_"+job_id
    
    # Set file lock path
    file_lock_path = config._ILP_LOCK_DIR+dataset_type+"_"+str(section_number)+"_rand" + ".lockfile"+"_"+job_id
    
    # If the graph is locked skip
    if check_lock(file_lock_path):
        print section_number
        print "File locked, skipping..."
        return
    
    # Check if already computed
    if check_lock(rand_output_path):
        print rand_output_path
        print "Already computed, skipping..."
        return
    
    # Set file lock so that no other processes attempt to build this graph 
    print "Setting file lock..."
    set_lock(file_lock_path)
    sys.stdout.flush()           
    
    # Load graph
    print "* Loading graph for Section # "+str(section_number)+" "+section_title+"..."
    sys.stdout.flush()
    try:
        print "* Loading from AFS..."
        g = utils.load_pickle(get_graph_pickle_path(graph_path, section_number, job_id))
        
    except:

        print "* Loading from web..."
        try:
            g = utils.load_web_pickle(get_graph_pickle_path(_SHREDDER_PATH, section_number, job_id))
        except:
            print "* Loading Failed skipping..."
        
            # Remove building file lock
            print "* Removing file lock..."
            remove_lock(file_lock_path)
            sys.stdout.flush()
            return
    
    sys.stdout.flush()

    # Number of content words that are both in the title and section text
    # number_of_content_words = int(sum(binary_stemmed_section_word_scores[section_number].values()))
    # Get number of content words in the title
    bigrams._STOP_WORD_FILTERING = True
    number_of_content_words = len(bigrams.split_into_words(section_title.lower()))
    # number_of_content_words = 3
    
    # No file locking
    # Run Rand Search
    print "* Running Randomized for "+str(num_max_iters)+"..."
    run_rand_iters(g, num_max_iters, number_of_content_words, rand_output_path, ilp_score)
    
    # Remove building file lock
    print "* Removing file lock..."
    remove_lock(file_lock_path)
    sys.stdout.flush()
    
    return

def build(graph_path, dataset_type, datafile, maxent_stub, section_titles, job_id, lm_smoothing, tag_smoothing):

    stemmed_section_word_scores = {}
    print "* Loading binary selection scores..."
    binary_stemmed_section_word_scores = lattice.load_binary_selection_scores(maxent_stub+".sectiontitles", maxent_stub+".testwords")
    
    print "* Loading section words stemmed with scores..."
    non_binary_stemmed_section_word_scores = lattice.get_stemmed_section_word_scores(maxent_stub+".maxent_out", maxent_stub+".testwords")
        
    if _BINARY_SELECTION_SCORES:
        stemmed_section_word_scores = binary_stemmed_section_word_scores
    else:
        stemmed_section_word_scores = non_binary_stemmed_section_word_scores
  
    section_numbers = sorted(stemmed_section_word_scores.keys())

    print "* Loading word to tag mapping..."
    word_to_tag_path = config._TERMINAL_PROBS_PICKLE_PATH+dataset_type+"_word_to_tag"+".pickle"
    word_to_tag = utils.load_pickle(word_to_tag_path)    

    print "* Loading tag probs..."
    if tag_smoothing == "true":
        print "* Loading smoothed tag lm..."
        training_sections = lattice.get_train_section_titles(maxent_stub)
        tag_lm = srilm_tagger.SRILM_tagger(dataset_type, datafile, training_sections)
    else:
        print "* Loading unsmoothed tag lm..."
        tag_lm_path = config._TERMINAL_PROBS_PICKLE_PATH+dataset_type+"_tag_lm"+".pickle"
        tag_lm = utils.load_pickle(tag_lm_path)    

    print "* Loading terminal word probs..."
    term_word_probs = terminal_word_probs.get_terminal_word_probs(datafile, dataset_type, maxent_stub)

    print "* Loading stop words..."
    (single_stopwords, bigram_stopwords, all_stopwords) = lattice.load_all_stopwords(config._STOPWORDS_FILE, datafile, dataset_type)
    
    print "* Loading language model..."
    if lm_smoothing == "true":
        print "* Loading SRILM language model..."
        lang_model = srilm_arpa.SRILM(dataset_type, datafile, bigram_stopwords)
    else:
        print "* Loading non-SRILM language model..."
        lang_model = language_model.load_language_model(dataset_type)

    print "* Loading destemming probabilities..."
    stem_probs = lattice.get_stemming_probs(datafile, dataset_type)
    
    # Loop through each section
    for section_number in section_numbers:

        section_title = section_titles[section_number]
        
        # Load or generate graph
        print "* Loading graph for Section # "+section_number+" "+section_title+"..."
        sys.stdout.flush()
        
        # If the graph already exists skip
        if check_lock(get_graph_pickle_path(graph_path, section_number, job_id)):
            print "Graph already exists, skipping..."
            continue

        # Set file lock path
        file_lock_path = config._GRAPH_BUILDING_LOCK_DIR+dataset_type+"_"+section_number+".lockfile"+"_"+job_id
        
        # If the graph is locked skip
        if check_lock(file_lock_path):
            print "File locked, skipping..."
            continue

        # Set file lock so that no other processes attempt to build this graph 
        set_lock(file_lock_path)
        print "Setting file lock..."
        sys.stdout.flush()
        
        # Build graph
        print "* Building graph..."
        sys.stdout.flush()
        g = graph.Graph(stemmed_section_word_scores[section_number], stem_probs, lang_model, single_stopwords, tag_lm, word_to_tag, term_word_probs, section_number)
        sys.stdout.flush()

        # Save graph
        print "* Saving graph to file..."
	graph_save_path = get_graph_pickle_path(graph_path, section_number, job_id)
	print graph_save_path
        utils.save_pickle(graph_save_path, g)
        sys.stdout.flush()

        # Remove building file lock
        print "* Removing file lock..."
        remove_lock(file_lock_path)
        sys.stdout.flush()

    return

def make_ilp(section_number, dataset, graph_path, section_title, formulation_dir, job_id):

    # Set formulation output path
    ilp_formulation_path = formulation_dir + dataset + "_" + str(section_number) + ".lpfile_"+job_id
    
    # Set file lock path
    file_lock_path = config._ILP_LOCK_DIR+dataset + "_" + str(section_number) + "_.lockfile_"+job_id
    
    # If the graph is locked skip
    if check_lock(file_lock_path):
        print "File locked, skipping..."
        return
    
    # Check if already computed
    if check_lock(ilp_formulation_path):
        print "Already computed, skipping..."
        return

    # Set file lock so that no other processes attempt to build this graph 
    print "Setting file lock..."
    set_lock(file_lock_path)
    sys.stdout.flush()                
    
    # Load graph
    print "* Loading graph for Section # " + str(section_number) + " " + section_title + "..."
    sys.stdout.flush()
    try:
        print get_graph_pickle_path(graph_path, section_number, job_id)
        g = utils.load_pickle(get_graph_pickle_path(graph_path, section_number, job_id))
    except:
        print "* Loading Failed skipping..."
        
        # Remove building file lock
        print "* Removing file lock..."
        remove_lock(file_lock_path)
        sys.stdout.flush()
        
        return
    sys.stdout.flush()
        
    # Number of content words that are both in the title and section text
    # number_of_content_words = int(sum(binary_stemmed_section_word_scores[section_number].values()))
    # Get number of content words in the title
    bigrams._STOP_WORD_FILTERING = True
    number_of_content_words = len(bigrams.split_into_words(section_title.lower()))
    
    # Formulate ILP constraints
    print "* Formulating ILP..."
    g.make_ilp_file(number_of_content_words+1, ilp_formulation_path)
    
    # Remove building file lock
    print "* Removing file lock..."
    remove_lock(file_lock_path)
    sys.stdout.flush()

def run_tagger_lm(dataset, datafile, maxent_stub):
    
    # Generate word to tag mappings
    # Try to load from pickle
    print "* Loading word to tag...."
    word_to_tag_path = config._TERMINAL_PROBS_PICKLE_PATH+dataset+"_word_to_tag"+".pickle"
    try:
        print " * Loading pickle..."
        print word_to_tag_path
        word_to_tag = utils.load_pickle(word_to_tag_path)
    except:
        print " * Rebuilding pickle..."
        test_section_numbers = lattice.get_test_section_numbers(maxent_stub)
        word_to_tag = tagger.WordToTag(dataset, datafile, test_section_numbers)
        utils.save_pickle(word_to_tag_path, word_to_tag)

    # Generate unsmoothed tag language model
    # Try to load from pickle
    print "* Loading unsmoothed tag lm..."
    tag_lm_path = config._TERMINAL_PROBS_PICKLE_PATH+dataset+"_tag_lm"+".pickle"
    training_sections = lattice.get_train_section_titles(maxent_stub)
    try:
        tag_lm = utils.load_pickle(tag_lm_path)
    except:
        tag_lm = tagger.TitleTagLanguageModel(datafile, dataset, training_sections)
        utils.save_pickle(tag_lm_path, tag_lm)

    # Generate smoothed tag language model
    print "* Loading smoothed tag lm..."
    srilm_tagger.SRILM_tagger(dataset, datafile, training_sections)

    # Generate unsmoothed bigram language model
    print "* Loading unsmoothed lm..."
    (single_stopwords, bigram_stopwords, all_stopwords) = lattice.load_all_stopwords(config._STOPWORDS_FILE, datafile, dataset)
    lang_model = language_model.LanguageModel(datafile, dataset, bigram_stopwords, single_stopwords)
    language_model.save_language_model(dataset, lang_model)

    # Generate smoothed bigram language model
    print "* Loading smoothed lm..."
    srilm_arpa.SRILM(dataset, datafile, bigram_stopwords)
    
def main():
    main_run()

def main_run():

    try:
        opts, args = getopt.getopt(sys.argv[1:], "", ["dataset=", "datafile=", "mode=", "maxent_stub=", "graph_dir=", "max_rand_iters=", "beam_size=", "sec_num=", "ilp_outfile=", "job_id=", "lm_smoothing=", "tag_smoothing=", "formulation_dir="])
    except getopt.GetoptError:
        print "Invalid arguments supplied"
        sys.exit(2)

    # Get command line arguments
    arg_datafile = None
    arg_dataset = None
    arg_maxent_stub = None
    arg_graph_dir = None
    arg_beam_size = None
    arg_mode = None
    arg_max_rand_iters = None
    arg_beam_size = None
    arg_sec_num = None    
    arg_ilp_outfile = None
    arg_job_id = None
    arg_lm_smoothing = None
    arg_tag_smoothing = None
    #arg_formulation_dir = None
    
    for opt, val in opts:

        if opt in "--mode":
            arg_mode = val

        elif opt in "--datafile":
            arg_datafile = val

        elif opt in "--dataset":
            arg_dataset = val

        elif opt in "--maxent_stub":
            arg_maxent_stub = val

        elif opt in "--graph_dir":
            arg_graph_dir = val

        elif opt in "--max_rand_iters":
            arg_max_rand_iters = int(val)

        elif opt in "--beam_size":
            arg_beam_size = int(val)

        elif opt in "--sec_num":
            arg_sec_num = int(val)

        elif opt in "--ilp_outfile":
            arg_ilp_outfile = val

        elif opt in "--job_id":
            arg_job_id = val

        elif opt in "--lm_smoothing":
            arg_lm_smoothing = val

        elif opt in "--tag_smoothing":
            arg_tag_smoothing = val

        #elif opt in "--formulation_dir":
        #    arg_formulation_dir = val
            
    assert(arg_mode != None and arg_datafile != None and arg_dataset != None)

    if arg_dataset == "algorithms_rnd_baseline": arg_dataset = "algorithms"
    
    # Load actual section titles
    section_titles = lattice.get_section_titles(arg_datafile, arg_dataset)

    if arg_mode == "beam":
        assert(arg_dataset != None and arg_beam_size != None and arg_graph_dir != None and arg_job_id != None and arg_maxent_stub != None)
        search_beam(section_titles[str(arg_sec_num)], arg_dataset, arg_beam_size, arg_graph_dir, arg_maxent_stub, arg_job_id)

    #elif arg_mode == "ilp":
    #    assert(arg_sec_num != None and arg_dataset != None and arg_graph_dir != None and arg_job_id != None and arg_formulation_dir != None)
    #    search_ilp(arg_sec_num, section_titles[str(arg_sec_num)], arg_dataset, arg_graph_dir, arg_formulation_dir, arg_job_id)

    elif arg_mode == "rand":
        assert(arg_dataset != None and arg_max_rand_iters != None \
               and arg_ilp_outfile != None and arg_graph_dir != None and arg_job_id != None and arg_maxent_stub != None)
        search_rand_loop( section_titles, arg_max_rand_iters, arg_dataset, arg_graph_dir, arg_ilp_outfile, arg_maxent_stub, arg_job_id)

    elif arg_mode == "build":
        assert(arg_datafile != None and arg_dataset != None and arg_graph_dir != None and arg_maxent_stub != None and \
               arg_job_id != None and arg_lm_smoothing != None and arg_tag_smoothing != None)
        build(arg_graph_dir, arg_dataset, arg_datafile, arg_maxent_stub, section_titles, arg_job_id, arg_lm_smoothing, arg_tag_smoothing)

    elif arg_mode == "taglm":
        assert(arg_datafile != None and arg_dataset != None and arg_maxent_stub != None)
        run_tagger_lm(arg_dataset, arg_datafile, arg_maxent_stub)

    #elif arg_mode == "makeilp":
    #    assert(arg_sec_num != None and arg_dataset != None and arg_graph_dir != None and arg_graph_dir != None and arg_formulation_dir != None)
    #    make_ilp(arg_sec_num, arg_dataset, arg_graph_dir, section_titles[str(arg_sec_num)], arg_formulation_dir, arg_job_id)

    else:
        raise "Error: Invalid Mode ("+arg_mode+")"

''' ngram -lm BaseLineAlgorithms.LM -debug 3 -ppl test '''

if __name__ == '__main__':
    sys.exit( main( ) )
